from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Sequence, TypeVar

import numpy as np
from cmap import Colormap
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers._mixin import (
    EdgeNamespace,
    FaceEdgeMixinEvents,
    FaceNamespace,
    MultiFaceEdgeMixin,
)
from whitecanvas.layers._sizehint import xyz_size_hint
from whitecanvas.layers.layer3d._base import DataBoundLayer3D
from whitecanvas.protocols import MarkersProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Symbol,
    XYZData,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xyz

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers._mixin import ConstEdge, ConstFace, MultiEdge, MultiFace

_void = _Void()
_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)
_Size = TypeVar("_Size", float, NDArray[np.floating])


class MarkersLayerEvents(FaceEdgeMixinEvents):
    clicked = Signal(list)
    symbol = Signal(Symbol)
    size = Signal(float)


class Markers3D(
    DataBoundLayer3D[MarkersProtocol, XYZData],
    MultiFaceEdgeMixin[_Face, _Edge],
    Generic[_Face, _Edge, _Size],
):
    events: MarkersLayerEvents
    _events_class = MarkersLayerEvents
    _backend_class_name = "components3d.Markers3D"

    if TYPE_CHECKING:

        def __new__(cls, *args, **kwargs) -> Markers3D[ConstFace, ConstEdge, float]: ...

    def __init__(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        zdata: ArrayLike1D,
        *,
        name: str | None = None,
        symbol: Symbol | str = Symbol.CIRCLE,
        size: float = 12.0,
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        MultiFaceEdgeMixin.__init__(self)
        xdata, ydata, zdata = normalize_xyz(xdata, ydata, zdata)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), xdata, ydata, zdata)
        self._size_is_array = False
        self.update(symbol=symbol, size=size, color=color, hatch=hatch, alpha=alpha)
        self.edge.color = color
        if not self.symbol.has_face():
            self.edge.update(width=2.0, color=color)
        pad_r = size / 400
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(
            xdata, ydata, zdata, pad_r, pad_r, pad_r
        )

        # self._backend._plt_connect_pick_event(self.events.clicked.emit)
        self._init_events()

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return self.data.x.size

    @classmethod
    def from_dict(
        cls, d: dict[str, Any], backend: Backend | str | None = None
    ) -> Markers3D:
        return cls(
            d["data"]["x"], d["data"]["y"], d["data"]["z"], name=d["name"],
            symbol=d["symbol"], size=d["size"], color=d["face"]["color"],
            hatch=d["face"]["hatch"], backend=backend,
        ).with_edge(
            color=d["edge"]["color"], width=d["edge"]["width"], style=d["edge"]["style"]
        )  # fmt: skip

    def to_dict(self) -> dict[str, Any]:
        return {
            "type": f"{self.__module__}.{self.__class__.__name__}",
            "data": self.data.to_dict(),
            "name": self.name,
            "symbol": self.symbol,
            "size": self.size,
            "face": self.face.to_dict(),
            "edge": self.edge.to_dict(),
        }

    def _get_layer_data(self) -> XYZData:
        return XYZData(*self._backend._plt_get_data())

    def _norm_layer_data(self, data: Any) -> XYZData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 3:
                raise ValueError("Expected a (N, 3) array")
            xdata, ydata, zdata = data[:, 0], data[:, 1], data[:, 2]
        else:
            xdata, ydata, zdata = data
        x0, y0, z0 = self.data
        if xdata is not None:
            x0 = as_array_1d(xdata)
        if ydata is not None:
            y0 = as_array_1d(ydata)
        if zdata is not None:
            z0 = as_array_1d(zdata)
        if x0.size != y0.size or x0.size != z0.size:
            raise ValueError(
                "Expected xdata, ydata and zdata to have the same size, "
                f"got {x0.size}, {y0.size} and {z0.size}."
            )
        return XYZData(x0, y0, z0)

    def _set_layer_data(self, data: XYZData):
        x0, y0, z0 = data
        self._backend._plt_set_data(x0, y0, z0)
        if self._size_is_array:
            pad_r = self.size.mean() / 400
        else:
            pad_r = self.size / 400
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(
            x0, y0, z0, pad_r, pad_r, pad_r
        )

    def set_data(
        self,
        xdata: ArrayLike | None = None,
        ydata: ArrayLike | None = None,
        zdata: ArrayLike | None = None,
    ):
        if xdata is None:
            xdata = self.data.x
        if ydata is None:
            ydata = self.data.y
        if zdata is None:
            zdata = self.data.z
        self.data = XYZData(xdata, ydata, zdata)

    @property
    def symbol(self) -> Symbol:
        """Symbol used to mark the data points."""
        return self._backend._plt_get_symbol()

    @symbol.setter
    def symbol(self, symbol: str | Symbol):
        sym = Symbol(symbol)
        if self.ndata > 0:
            self._backend._plt_set_symbol(sym)
        self.events.symbol.emit(sym)

    @property
    def size(self) -> _Size:
        """Size of the symbol."""
        size = self._backend._plt_get_symbol_size()
        if self._size_is_array:
            return size
        elif size.size > 0:
            return size[0]
        else:
            return 10.0  # default size

    @size.setter
    def size(self, size: _Size):
        """Set marker size"""
        ndata = self.ndata
        if not isinstance(size, (float, int, np.number)):
            if not self._size_is_array:
                raise ValueError(
                    "Expected size to be a scalar. Use with_size_multi() to "
                    "set multiple sizes."
                )
            size = as_array_1d(size)
            if size.size != ndata:
                raise ValueError(
                    f"Expected `size` to have the same size as the layer data size "
                    f"({self.ndata}), got {size.size}."
                )
        if ndata > 0:
            self._backend._plt_set_symbol_size(size)
        self.events.size.emit(size)

    def update(
        self,
        *,
        symbol: Symbol | str | _Void = _void,
        size: float | _Void = _void,
        color: ColorType | _Void = _void,
        alpha: float | _Void = _void,
        hatch: str | Hatch | _Void = _void,
    ) -> Self:
        """Update the properties of the markers."""
        if symbol is not _void:
            self.symbol = symbol
        if size is not _void:
            self.size = size
        if color is not _void:
            self.face.color = color
        if hatch is not _void:
            self.face.hatch = hatch
        if alpha is not _void:
            self.face.alpha = alpha
        return self

    def color_by_density(
        self,
        cmap: ColormapType = "jet",
        *,
        width: float = 0.0,
    ) -> Self:
        """
        Set the color of the markers by density.

        Parameters
        ----------
        cmap : ColormapType, optional
            Colormap used to map the density to colors.
        """
        from whitecanvas.utils.kde import gaussian_kde

        xyzdata = self.data
        xyz = np.vstack([xyzdata.x, xyzdata.y, xyzdata.z])
        density = gaussian_kde(xyz)(xyz)
        normed = density / density.max()
        self.with_face_multi(color=Colormap(cmap)(normed))
        if width is not None:
            self.width = width
        return self

    def as_edge_only(
        self,
        *,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the markers to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        color = self.face.color
        if color.ndim == 0:
            pass
        elif color.ndim == 1:
            self.with_edge(color=color, width=width, style=style)
        elif color.ndim == 2:
            self.with_edge_multi(color=color, width=width, style=style)
        else:
            raise RuntimeError("Unreachable error.")
        self.face.update(alpha=0.0)
        return self

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1.0,
    ) -> Markers3D[ConstFace, _Edge, _Size]:
        """
        Return a markers layer with constant face properties.

        In most cases, this function is not needed because the face properties
        can be set directly on construction. Examples below are equivalent:

        >>> markers = canvas.add_markers(x, y).with_face(color="red")
        >>> markers = canvas.add_markers(x, y, color="red")

        This function will be useful when you want to change the markers from
        multi-face state to constant-face state:

        >>> markers = canvas.add_markers(x, y).with_face_multi(color=colors)
        >>> markers = markers.with_face(color="red")

        Parameters
        ----------
        color : color-like, optional
            Color of the marker faces.
        hatch : str or FacePattern, optional
            Hatch pattern of the faces.
        alpha : float, optional
            Alpha channel of the faces.

        Returns
        -------
        Markers3D
            The updated markers layer.
        """
        super().with_face(color, hatch, alpha)
        if not self.symbol.has_face():
            width = self.edge.width
            if isinstance(width, (float, int, np.number)):
                self.edge.update(width=width or 1.0, color=color)
            else:
                self.edge.update(width=width[0] or 1.0, color=color)
        return self

    def with_face_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] | _Void = _void,
        alpha: float = 1,
    ) -> Markers3D[MultiFace, _Edge, _Size]:
        """
        Return a markers layer with multi-face properties.

        This function is used to create a markers layer with multiple face
        properties, such as colorful markers.

        >>> markers = canvas.add_markers(x, y).with_face_multi(color=colors)

        Parameters
        ----------
        color : color-like or sequence of color-like, optional
            Color(s) of the marker faces.
        hatch : str or Hatch or sequence of it, optional
            Pattern(s) of the faces.
        alpha : float or sequence of float, optional
            Alpha channel(s) of the faces.

        Returns
        -------
        Markers3D
            The updated markers layer.
        """
        super().with_face_multi(color, hatch, alpha)
        if not self.symbol.has_face():
            width = self.edge.width
            if isinstance(width, (float, int, np.number)):
                self.edge.update(width=width or 1.0, color=color)
            else:
                self.edge.update(width=width[0] or 1.0, color=color)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Markers3D[_Face, ConstEdge, _Size]:
        return super().with_edge(color, width, style, alpha)

    def with_edge_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1.0,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Markers3D[_Face, MultiEdge, _Size]:
        """
        Return a markers layer with multi-edge properties.

        This function is used to create a markers layer with multiple edge
        properties, such as colorful markers.

        >>> markers = canvas.add_markers(x, y).with_edge_multi(color=colors)

        Parameters
        ----------
        color : color-like or sequence of color-like, optional
            Color(s) of the marker faces.
        width : float or array of float, optional
            Width(s) of the edges.
        style : str, LineStyle or sequence of them, optional
            Line style(s) of the edges.
        alpha : float or sequence of float, optional
            Alpha channel(s) of the faces.

        Returns
        -------
        Markers3D
            The updated markers layer.
        """
        return super().with_edge_multi(color, width, style, alpha)

    def with_size_multi(
        self,
        size: float | Sequence[float],
    ) -> Markers3D[_Face, _Edge, NDArray[np.float32]]:
        if isinstance(size, (float, int, np.number)):
            _size = np.full(self.ndata, size, dtype=np.float32)
        else:
            _size = as_array_1d(size, dtype=np.float32)
            if _size.size != self.ndata:
                raise ValueError(
                    "Expected size to have the same size as the data, "
                    f"got {_size.size} and {self.ndata}"
                )
        self._size_is_array = True
        self.size = _size
        return self



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from whitecanvas.backend import Backend
    from whitecanvas.layers._base import Layer

_CLASS_MAP: dict[str, type[Layer]] = {}


def _pick_layer_class(name: str) -> type[Layer]:
    import importlib

    if name not in _CLASS_MAP:
        mod_name, cls_name = name.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        _CLASS_MAP[name] = getattr(mod, cls_name)
    return _CLASS_MAP[name]


def construct_layer(d: dict[str, Any], backend: Backend | str | None = None) -> Layer:
    """Construct a layer from a dictionary."""
    if "type" not in d:
        raise ValueError(f"Layer dict must have a 'type' key, got {d!r}.")
    typ = d["type"]
    if not isinstance(typ, str):
        raise ValueError(f"Layer type must be a string, got {typ!r}.")
    return _pick_layer_class(typ).from_dict(d, backend=backend)


def construct_layers(
    d: list[dict[str, Any]],
    backend: Backend | str | None = None,
) -> list[Layer]:
    """Construct a list of layers from a list of dictionaries."""
    return [construct_layer(c, backend=backend) for c in d]



from __future__ import annotations

import json
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any

import numpy as np
from cmap import Color, Colormap

from whitecanvas.canvas._palette import ColorPalette


@lru_cache(maxsize=1)
def catalog_names():
    return set(Colormap.catalog().keys())


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, dict):
            if isinstance(color := obj.get("color"), np.ndarray):
                if color.ndim == 1:
                    return Color(color).hex
                else:
                    return [Color(c).hex for c in color]
            else:
                return super().default(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, Path):
            return str(obj)
        elif isinstance(obj, Color):
            return obj.hex
        elif isinstance(obj, Colormap):
            if obj.name in catalog_names():
                return obj.name
            return [(float(k), v.hex) for k, v in obj.color_stops]
        elif isinstance(obj, ColorPalette):
            return self.default(obj._cmap)
        elif hasattr(obj, "to_dict"):
            return color_to_hex(obj.to_dict())
        elif callable(obj):
            raise TypeError(f"Function {obj!r} is not JSON serializable")
        return super().default(obj)


def color_to_hex(d: dict[str, Any]) -> dict[str, Any]:
    to_update = {}
    for k, v in d.items():
        if isinstance(v, dict):
            to_update[k] = color_to_hex(v)
        elif isinstance(v, (tuple, list)):
            to_update[k] = [color_to_hex(e) if isinstance(e, dict) else e for e in v]
        elif k == "color" and isinstance(v, np.ndarray):
            if v.ndim == 1:
                to_update[k] = Color(v).hex
            else:
                to_update[k] = [Color(c).hex for c in v]
    d.update(to_update)
    return d



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Sequence, TypeVar

_T = TypeVar("_T")


class _SerializableSorter(ABC, Generic[_T]):
    @abstractmethod
    def __call__(self, x: Sequence[_T]) -> Sequence[_T]: ...
    @classmethod
    @abstractmethod
    def from_dict(cls, d: dict[str, str]) -> _SerializableSorter[_T]: ...
    @abstractmethod
    def to_dict(self) -> dict[str, str]: ...


class IdentitySorter(_SerializableSorter[_T]):
    def __call__(self, x: Sequence[_T]) -> Sequence[_T]:
        return x

    @classmethod
    def from_dict(cls, *_) -> IdentitySorter[_T]:
        return cls()

    def to_dict(self) -> dict[str, str]:
        return {"type": "identity"}


class SimpleSorter(_SerializableSorter[_T]):
    def __init__(self, ascending: bool) -> None:
        self._ascending = ascending

    def __call__(self, x: Sequence[_T]) -> Sequence[_T]:
        return sorted(x, reverse=not self._ascending)

    @classmethod
    def from_dict(cls, d: dict[str, str]) -> SimpleSorter[_T]:
        return cls(d["type"] == "ascending")

    def to_dict(self) -> dict[str, str]:
        return {"type": "ascending" if self._ascending else "descending"}


class ManualSorter(_SerializableSorter[_T]):
    def __init__(self, order: Sequence[_T]) -> None:
        if len(order) != len(set(order)):
            raise ValueError(f"Order contains duplicate values: {order!r}")
        self._order = list(order)

    @classmethod
    def norm_tuple(
        cls,
        order: Sequence[tuple[_T, ...] | _T],
    ) -> ManualSorter[tuple[_T, ...]]:
        order_normed = [u if isinstance(u, tuple) else (u,) for u in order]
        return cls(order_normed)

    def __call__(self, x: Sequence[_T]) -> Sequence[_T]:
        input_categories = set(x)
        for each in self._order:
            if each in input_categories:
                input_categories.discard(each)
        if input_categories:
            raise ValueError(
                f"Categories {input_categories!r} are not in the given order "
                f"{self._order!r}"
            )
        return self._order

    @classmethod
    def from_dict(cls, d: dict[str, str]) -> ManualSorter[_T]:
        return cls(d["order"])

    def to_dict(self) -> dict[str, str]:
        return {"type": "manual", "order": self._order}


def construct_sorter(d: dict[str, str]) -> _SerializableSorter[_T]:
    if d["type"] == "identity":
        return IdentitySorter.from_dict(d)
    elif d["type"] == "ascending" or d["type"] == "descending":
        return SimpleSorter.from_dict(d)
    elif d["type"] == "manual":
        return ManualSorter.from_dict(d)
    else:
        raise ValueError(f"Unknown sorter type: {d['type']!r}")



from __future__ import annotations

from typing import Any

import numpy as np

from whitecanvas.utils.normalize import as_array_1d
from whitecanvas.utils.type_check import is_real_number


class TextOffset:
    def _asarray(self) -> tuple[Any, Any]:
        raise NotImplementedError

    def _add(self, dx, dy) -> TextOffset:
        raise NotImplementedError


class NoOffset(TextOffset):
    def _asarray(self):
        return (0, 0)

    def _add(self, dx, dy) -> TextOffset:
        if is_real_number(dx) and is_real_number(dy):
            if dx == 0 and dy == 0:
                return self
            return ConstantOffset(dx, dy)
        else:
            return CustomOffset(as_array_1d(dx), as_array_1d(dy))

    def __repr__(self) -> str:
        return "<NoOffset>"

    @classmethod
    def from_dict(cls, _: dict[str, Any]) -> NoOffset:
        return cls()

    def to_dict(self) -> dict[str, Any]:
        return {"type": "no"}


class ConstantOffset(TextOffset):
    def __init__(self, x: float, y: float):
        self._x, self._y = x, y

    def _asarray(self) -> tuple[float, float]:
        return (self._x, self._y)

    def _add(self, dx, dy) -> TextOffset:
        if is_real_number(dx) and is_real_number(dy):
            return ConstantOffset(self._x + dx, self._y + dy)
        else:
            return CustomOffset(as_array_1d(dx) + self._x, as_array_1d(dy) + self._y)

    def __repr__(self) -> str:
        return f"<ConstantOffset({self._x}, {self._y})>"

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> ConstantOffset:
        return cls(d["x"], d["y"])

    def to_dict(self) -> dict[str, Any]:
        return {"type": "constant", "x": self._x, "y": self._y}


class CustomOffset(TextOffset):
    def __init__(self, x: Any, y: Any):
        self._x, self._y = x, y

    def _asarray(self) -> tuple[Any, Any]:
        return (self._x, self._y)

    def _add(self, dx, dy) -> TextOffset:
        return CustomOffset(as_array_1d(dx + self._x), as_array_1d(dy + self._y))

    def __repr__(self) -> str:
        return f"<CustomOffset({self._x}, {self._y})>"

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> CustomOffset:
        return cls(
            np.asarray(d["x"], dtype=np.float32), np.asarray(d["y"], dtype=np.float32)
        )

    def to_dict(self) -> dict[str, Any]:
        return {"type": "custom", "x": self._x, "y": self._y}


def parse_offset_dict(d: dict[str, Any]) -> TextOffset:
    if d["type"] == "no":
        return NoOffset.from_dict(d)
    elif d["type"] == "constant":
        return ConstantOffset.from_dict(d)
    elif d["type"] == "custom":
        return CustomOffset.from_dict(d)
    else:
        raise ValueError(f"Unknown offset type: {d['type']}")



from __future__ import annotations

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray


class XYData(NamedTuple):
    """
    Tuple of x and y array.

    Used for data of Line, Markers etc.
    """

    x: NDArray[np.floating]
    y: NDArray[np.floating]

    def stack(self) -> NDArray[np.floating]:
        """Data as a stacked (N, 2) array."""
        return np.stack([self.x, self.y], axis=1)

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYData:
        """Create XYData from a dictionary."""
        return cls(data["x"], data["y"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"x": self.x, "y": self.y}


class XYYData(NamedTuple):
    """
    Tuple of x, y0, and y1 array.

    Used for data of Bars, Errorbars etc.
    """

    x: NDArray[np.floating]
    y0: NDArray[np.floating]
    y1: NDArray[np.floating]

    @property
    def ycenter(self) -> NDArray[np.floating]:
        return (self.y0 + self.y1) / 2

    @property
    def ydiff(self) -> NDArray[np.floating]:
        return self.y1 - self.y0

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYYData:
        """Create XYYData from a dictionary."""
        return cls(data["x"], data["y0"], data["y1"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"x": self.x, "y0": self.y0, "y1": self.y1}


class XYTextData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    text: NDArray[np.object_]

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYTextData:
        """Create XYTextData from a dictionary."""
        return cls(data["x"], data["y"], data["text"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"x": self.x, "y": self.y, "text": self.text}


class XYZData(NamedTuple):
    """
    Tuple of x, y, and z array.

    Used for data of Surface etc.
    """

    x: NDArray[np.floating]
    y: NDArray[np.floating]
    z: NDArray[np.floating]

    def stack(self) -> NDArray[np.floating]:
        """Data as a stacked (N, 3) array."""
        return np.stack([self.x, self.y, self.z], axis=1)

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYZData:
        """Create XYZData from a dictionary."""
        return cls(data["x"], data["y"], data["z"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"x": self.x, "y": self.y, "z": self.z}


class XYVectorData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    vx: NDArray[np.floating]
    vy: NDArray[np.floating]

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYVectorData:
        """Create XYVectorData from a dictionary."""
        return cls(data["x"], data["y"], data["vx"], data["vy"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"x": self.x, "y": self.y, "vx": self.vx, "vy": self.vy}


class XYZVectorData(NamedTuple):
    x: NDArray[np.floating]
    y: NDArray[np.floating]
    z: NDArray[np.floating]
    vx: NDArray[np.floating]
    vy: NDArray[np.floating]
    vz: NDArray[np.floating]

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> XYZVectorData:
        """Create XYZVectorData from a dictionary."""
        return cls(data["x"], data["y"], data["z"], data["vx"], data["vy"], data["vz"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {
            "x": self.x, "y": self.y, "z": self.z,
            "vx": self.vx, "vy": self.vy, "vz": self.vz
        }  # fmt: skip


class MeshData(NamedTuple):
    vertices: NDArray[np.floating]
    faces: NDArray[np.intp]

    @classmethod
    def from_dict(cls, data: dict[str, NDArray[np.floating]]) -> MeshData:
        """Create MeshData from a dictionary."""
        return cls(data["vertices"], data["faces"])

    def to_dict(self) -> dict[str, NDArray[np.floating]]:
        """Data as a dictionary."""
        return {"vertices": self.vertices, "faces": self.faces}


class Rect(NamedTuple):
    """Rectangular range in left, right, bottom, top order."""

    left: float
    right: float
    bottom: float
    top: float

    def __repr__(self) -> str:
        return (
            f"Rect(left={self.left:.6g}, right={self.right:.6g}, "
            f"bottom={self.bottom:.6g}, top={self.top:.6g})"
        )

    @classmethod
    def with_check(cls, left: float, right: float, bottom: float, top: float):
        if left > right:
            raise ValueError("left must be less than or equal to right")
        if bottom > top:
            raise ValueError("bottom must be less than or equal to top")
        return cls(float(left), float(right), float(bottom), float(top))

    @classmethod
    def with_sort(cls, left: float, right: float, bottom: float, top: float):
        left, right = sorted([left, right])
        bottom, top = sorted([bottom, top])
        return cls(float(left), float(right), float(bottom), float(top))

    @property
    def width(self) -> float:
        """Width of the range."""
        return self.right - self.left

    @property
    def height(self) -> float:
        """Height of the range."""
        return self.top - self.bottom

    @property
    def size(self) -> tuple[float, float]:
        """Size (width, height) of the range."""
        return self.width, self.height

    @property
    def center(self) -> tuple[float, float]:
        """Center of the range."""
        return (self.left + self.right) / 2, (self.top + self.bottom) / 2

    @classmethod
    def from_dict(cls, data: dict[str, float]) -> Rect:
        """Create Rect from a dictionary."""
        return cls(data["left"], data["right"], data["bottom"], data["top"])

    def to_dict(self) -> dict[str, float]:
        """Data as a dictionary."""
        return {
            "left": self.left,
            "right": self.right,
            "bottom": self.bottom,
            "top": self.top,
        }



from __future__ import annotations

import base64
from typing import Any, Mapping

import numpy as np
from cmap import Color
from numpy.typing import ArrayLike, NDArray

from whitecanvas.types import XYData, XYZData
from whitecanvas.utils import type_check as _tc


def as_array_1d(x: ArrayLike, dtype=None) -> NDArray[np.number]:
    """Normalize the input as a 1D array."""
    x = np.asarray(x, dtype=dtype)
    if x.ndim != 1:
        raise ValueError(f"Expected 1D array, got {x.ndim}D array")
    if x.dtype.kind not in "iuf":
        raise ValueError(f"Input {x!r} did not return a numeric array")
    return x


def normalize_xy(*args) -> tuple[NDArray[np.number], NDArray[np.number]]:
    """Normalize the input as two 1D array with the same shape."""
    if len(args) == 1:
        if isinstance(args[0], XYData):
            return args[0].x, args[0].y
        arr = np.asarray(args[0])
        if arr.dtype.kind not in "iuf":
            raise ValueError(f"Input {args[0]!r} did not return a numeric array")
        if arr.ndim == 1:
            ydata = arr
            xdata = np.arange(ydata.size)
        elif arr.ndim == 2 and arr.shape[1] == 2:
            xdata = arr[:, 0]
            ydata = arr[:, 1]
        else:
            raise ValueError(
                f"Expected 1D array or 2D array with shape (N, 2), got {arr.shape}"
            )
    elif len(args) == 2:
        if np.isscalar(args[0]) and np.isscalar(args[1]):
            return np.array(args[0]), np.array(args[1])
        xdata = as_array_1d(args[0])
        if not hasattr(args[1], "__array__") and callable(args[1]):
            ydata = args[1](xdata)
        else:
            ydata = as_array_1d(args[1])
        if xdata.size != ydata.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {xdata.size} and {ydata.size}"
            )
    else:
        raise TypeError(f"Expected 1 or 2 positional arguments, got {len(args)}")
    return xdata, ydata


def normalize_xyz(
    *args,
) -> tuple[NDArray[np.number], NDArray[np.number], NDArray[np.number]]:
    """Normalize the input as three 1D array with the same shape."""
    if len(args) == 1:
        if isinstance(args[0], XYZData):
            return args[0].x, args[0].y, args[0].z
        arr = np.asarray(args[0])
        if arr.dtype.kind not in "iuf":
            raise ValueError(f"Input {args[0]!r} did not return a numeric array")
        if arr.ndim == 2 and arr.shape[1] == 3:
            xdata = arr[:, 0]
            ydata = arr[:, 1]
            zdata = arr[:, 2]
        else:
            raise ValueError(
                f"Expected 1D array or 2D array with shape (N, 3), got {arr.shape}"
            )
    elif len(args) == 3:
        xdata = as_array_1d(args[0])
        ydata = as_array_1d(args[1])
        zdata = as_array_1d(args[2])
        if xdata.size != ydata.size != zdata.size:
            raise ValueError(
                "Expected xdata, ydata, and zdata to have the same size, "
                f"got {xdata.size}, {ydata.size}, and {zdata.size}"
            )
    else:
        raise TypeError(f"Expected 1 or 3 positional arguments, got {len(args)}")
    return xdata, ydata, zdata


def arr_color(color) -> np.ndarray:
    """Normalize a color input to a 4-element float array."""
    try:
        c = Color(color)
    except Exception:
        raise ValueError(f"Invalid input for a color: {color!r}") from None
    return np.array(c.rgba, dtype=np.float32)


def hex_color(color) -> str:
    """Normalize a color input to a #RRGGBBAA string."""
    return Color(color).hex


def rgba_str_color(color) -> str:
    """Normalize a color input to a rgba(r, g, b, a) string."""
    return Color(color).rgba_string


def as_any_1d_array(x: Any, size: int, dtype=None) -> np.ndarray:
    if _tc.is_not_array(x):
        out = np.full((size,), x, dtype=dtype)
    else:
        out = np.asarray(x, dtype=dtype)
        if out.shape != (size,):
            raise ValueError(f"Expected shape ({size},), got {out.shape}")
    return out


def as_color_array(color, size: int) -> NDArray[np.float32]:
    if isinstance(color, str):  # e.g. color = "black"
        col = arr_color(color)
        return np.repeat(col[np.newaxis, :], size, axis=0)
    if isinstance(color, np.ndarray):
        if color.size == 0 and size == 0:
            return color
        if color.dtype.kind in "OU":
            if color.shape != (size,):
                raise ValueError(
                    f"Expected color array of shape ({size},), got {color.shape}"
                )
            return np.stack([arr_color(each) for each in color], axis=0)
        elif color.shape in [(3,), (4,)]:
            col = arr_color(color)
            return np.repeat(col[np.newaxis, :], size, axis=0)
        elif color.shape in [(size, 3), (size, 4)]:
            return color
        else:
            if isinstance(color, np.ndarray):
                color_repr = f"ndarray of shape={color.shape} and dtype={color.dtype}"
            else:
                color_repr = repr(color)
            raise ValueError(
                f"Color array must have shape (3,), (4,), (N={size}, 3), or (N={size},"
                f" 4) but got {color_repr}."
            )
    if isinstance(color, Color):
        return as_color_array(np.fromiter(color, dtype=np.float32), size)
    try:
        arr = np.array(color)
    except ValueError as e:
        if str(e).startswith("setting an array element with a sequence"):
            # this happens when ["red", [0.0, 1.0, 0.0, 1.0]] is given
            color_normed = np.stack(
                [Color(c).rgba for c in color],
                axis=0,
                dtype=np.float32,
            )
            return color_normed
        else:
            raise
    return as_color_array(arr, size)


def parse_texts(template: str, ndata: int, extra: Any | None = None) -> dict[str, Any]:
    """Parse a template string and return parameters of hover texts."""
    params = {}
    if extra is None:
        extra = {}
    elif isinstance(extra, Mapping):
        pass
    elif _tc.is_pandas_dataframe(extra):
        return parse_texts(template, ndata, extra=extra.to_dict(orient="list"))
    elif _tc.is_polars_dataframe(extra):
        return parse_texts(template, ndata, extra=extra.to_dict())
    else:
        raise TypeError("extra must be a mapping.")
    for k, v in extra.items():
        if "{" + k not in template:
            continue
        if _tc.is_not_array(v):
            _v = np.full(ndata, v)
        else:
            _v = as_any_1d_array(v, ndata)
        if _v.size != ndata:
            raise ValueError(f"Expected {_v.size} elements, got {ndata} for {k!r}.")
        params[k] = v
    return params


def encode_array(arr: np.ndarray, key: str) -> dict[str, Any]:
    """Encode nD array."""
    return {
        key: base64.b64encode(arr).decode("ascii"),
        "shape": arr.shape,
        "dtype": arr.dtype.str,
    }


def decode_array(data: dict, key: str):
    """Decode base64 array if it is dict."""
    if isinstance(data, dict):
        data = np.frombuffer(base64.b64decode(data[key]), dtype=data["dtype"]).reshape(
            data["shape"]
        )
    return np.array(data)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, SupportsIndex, TypeVar, Union

_T = TypeVar("_T")


class DimAxis(ABC, Generic[_T]):
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        """Name of the axis."""
        return self._name

    @abstractmethod
    def current_index(self) -> SupportsIndex:
        """Current int index of the axis."""

    @abstractmethod
    def current_value(self) -> _T:
        """Current value of the axis."""

    @abstractmethod
    def set_index(self, idx: int) -> None:
        """Set current index to the axis."""

    @abstractmethod
    def set_value(self, value: _T):
        """Set current value of the axis."""


class RangeAxis(DimAxis[int]):
    """An axis with a sequential coordinate."""

    def __init__(self, name: str, size: int):
        super().__init__(name)
        self._size = size
        self._value = 0

    def __repr__(self) -> str:
        return f"RangeAxis(name={self.name}, size={self.size})"

    def current_index(self) -> SupportsIndex:
        return self._value

    def current_value(self) -> Any:
        return self._value

    def set_index(self, idx: int):
        if not 0 <= idx < self._size:
            raise ValueError(f"Size of axis {self!r} is {self._size} but got {idx!r}.")
        self._value = idx

    def set_value(self, value: int):
        return self.set_index(int(value))

    @property
    def size(self) -> int:
        return self._size

    def set_size(self, size: int):
        self._size = size
        self._value = min(self._value, size - 1)


Category = tuple[Union[int, str], ...]


class CategoricalAxis(DimAxis[Category]):
    """An axis with a categorical coordinate."""

    def __init__(self, name: str, categories: list[Category]):
        super().__init__(name)
        self._categories = list(categories)
        self._mapper = {c: i for i, c in enumerate(categories)}
        self._value = self._categories[0]

    def __repr__(self) -> str:
        return f"CategoricalAxis(name={self.name}, categories={self.categories})"

    def current_index(self) -> SupportsIndex:
        return self._mapper[self._value]

    def current_value(self) -> Category:
        return self._value

    def set_index(self, idx: int) -> None:
        return self.set_value(self._categories[idx])

    def set_value(self, value: Category):
        if value not in self._mapper:
            raise ValueError(
                f"Value must be one of {list(self._mapper)} but got {value!r}."
            )
        self._value = value

    @property
    def categories(self) -> list[Category]:
        """List of categories of this axis."""
        return self._categories.copy()



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, overload

import numpy as np
from cmap import Color
from numpy.typing import NDArray
from psygnal import Signal, SignalGroup
from typing_extensions import override

from whitecanvas import protocols
from whitecanvas._json_utils import CustomEncoder, color_to_hex
from whitecanvas.backend import Backend
from whitecanvas.canvas import Canvas, CanvasBase
from whitecanvas.canvas._linker import link_axes
from whitecanvas.layers._deserialize import construct_layers
from whitecanvas.theme import get_theme
from whitecanvas.utils.normalize import arr_color

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.types import ColormapType


class GridEvents(SignalGroup):
    drawn = Signal()


class _Serializable(ABC):
    @abstractmethod
    def to_dict(self) -> dict[str, Any]: ...
    @classmethod
    @abstractmethod
    def from_dict(
        cls, d: dict[str, Any], backend: Backend | str | None = None
    ) -> Self: ...

    @classmethod
    def read_json(
        cls,
        path_or_str: str | Path,
        *,
        backend: Backend | str | None = None,
    ) -> Self:
        """
        Construct a canvas grid from a JSON file or string.

        Parameters
        ----------
        path_or_str : str or Path
            The path to the JSON file or the JSON string.
        backend : Backend or str, optional
            The backend of the canvas.
        """

        if isinstance(path_or_str, Path):
            txt = path_or_str.read_text()
        elif not isinstance(path_or_str, str):
            raise TypeError(f"Expected a path or a string, got {path_or_str!r}")
        else:
            if "{" in path_or_str:
                txt = path_or_str
            else:
                txt = Path(path_or_str).read_text()
        return cls.from_dict(json.loads(txt), backend=backend)

    @overload
    def write_json(self) -> str: ...
    @overload
    def write_json(self, path: str | Path) -> None: ...

    def write_json(self, path: str | Path | None = None) -> str | None:
        """Return a JSON string or write to a file."""

        txt = json.dumps(color_to_hex(self.to_dict()), indent=2, cls=CustomEncoder)
        if path is None:
            return txt
        if isinstance(path, str):
            path = Path(path)
        path.write_text(txt)
        return None


class CanvasGrid(_Serializable):
    _CURRENT_INSTANCE: CanvasGrid | None = None
    events: GridEvents

    def __init__(
        self,
        heights: list[int],
        widths: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> None:
        self._heights = heights
        self._widths = widths
        self._backend = Backend(backend)
        self._backend_object = self._create_backend()
        self._canvas_array = np.empty((len(heights), len(widths)), dtype=object)
        self._canvas_array.fill(None)

        # link axes
        self._x_linked = False
        self._y_linked = False
        self._x_linker_ref = None
        self._y_linker_ref = None

        # update settings
        theme = get_theme()
        self.background_color = theme.background_color
        self.size = theme.canvas_size
        self.events = GridEvents()
        self.__class__._CURRENT_INSTANCE = self

    @property
    def shape(self) -> tuple[int, int]:
        """The (row, col) shape of the grid"""
        return self._canvas_array.shape

    def link_x(self, *, future: bool = True, hide_ticks: bool = True) -> Self:
        """
        Link all the x-axes of the canvases in the grid.

        >>> from whitecanvas import new_grid
        >>> g = new_grid(2, 2).link_x()  # link x-axes of all canvases

        Parameters
        ----------
        future : bool, default True
            If Ture, all the canvases added in the future will also be linked. Only link
            the existing canvases if False.
        """
        if self._x_linker_ref is not None:
            self._x_linker_ref.unlink_all()  # initialize linker
        to_link = []
        for (_r, _), _canvas in self._iter_canvas():
            to_link.append(_canvas.x)
            if hide_ticks and _r != self.shape[0] - 1:
                _canvas.x.ticks.visible = False
        self._x_linker_ref = link_axes(to_link)
        if future:
            self._x_linked = True
            if hide_ticks:
                self._backend_object._plt_set_spacings(6, 6)
        return self

    def link_y(self, *, future: bool = True, hide_ticks: bool = True) -> Self:
        """
        Link all the y-axes of the canvases in the grid.

        >>> from whitecanvas import new_grid
        >>> g = new_grid(2, 2).link_y()  # link y-axes of all canvases

        Parameters
        ----------
        future : bool, default True
            If Ture, all the canvases added in the future will also be linked. Only link
            the existing canvases if False.
        """
        if self._y_linker_ref is not None:
            self._y_linker_ref.unlink_all()
        to_link = []
        for (_, _c), _canvas in self._iter_canvas():
            to_link.append(_canvas.y)
            if hide_ticks and _c != 0:
                _canvas.y.ticks.visible = False
        self._y_linker_ref = link_axes(to_link)
        if future:
            self._y_linked = True
            if hide_ticks:
                self._backend_object._plt_set_spacings(6, 6)
        return self

    def __repr__(self) -> str:
        cname = type(self).__name__
        w, h = self._size
        hex_id = hex(id(self))
        return f"<{cname} ({w:.1f} x {h:.1f}) at {hex_id}>"

    def __getitem__(self, key: tuple[int, int]) -> Canvas:
        canvas = self._canvas_array[key]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        elif isinstance(canvas, np.ndarray):
            raise ValueError(f"Cannot index by {key}.")
        return canvas

    def _create_backend(self) -> protocols.CanvasGridProtocol:
        return self._backend.get("CanvasGrid")(
            self._heights, self._widths, self._backend._app
        )

    def fill(self, palette: ColormapType | None = None) -> Self:
        """Fill the grid with canvases."""
        for _ in self._iter_add_canvas(palette=palette):
            pass
        return self

    def add_canvas(
        self,
        row: int,
        col: int,
        rowspan: int = 1,
        colspan: int = 1,
        *,
        palette: str | None = None,
    ) -> Canvas:
        """Add a canvas to the grid at the given position"""
        for idx, item in np.ndenumerate(self._canvas_array[row, col]):
            if item is not None:
                raise ValueError(f"Canvas already exists at {idx}")
        backend_canvas = self._backend_object._plt_add_canvas(
            row, col, rowspan, colspan
        )
        canvas = self._canvas_array[row, col] = Canvas.from_backend(
            backend_canvas,
            backend=self._backend,
            palette=palette,
        )
        # Now backend axes/viewbox are created, we can install mouse events
        canvas._install_mouse_events()

        # link axes if needed
        if self._x_linked:
            self._x_linker_ref.link(canvas.x)
        if self._y_linked:
            self._y_linker_ref.link(canvas.y)
        canvas.events.drawn.connect(self.events.drawn.emit, unique=True, max_args=None)
        return canvas

    def add_canvas_3d(
        self,
        row: int,
        col: int,
        rowspan: int = 1,
        colspan: int = 1,
        *,
        palette: str | None = None,
    ):
        """Add a canvas to the grid at the given position"""
        from whitecanvas.canvas.canvas3d._base import Canvas3D

        for idx, item in np.ndenumerate(self._canvas_array[row, col]):
            if item is not None:
                raise ValueError(f"Canvas already exists at {idx}")
        backend_canvas = self._backend_object._plt_add_canvas_3d(
            row, col, rowspan, colspan
        )
        canvas = self._canvas_array[row, col] = Canvas3D.from_backend(
            backend_canvas,
            backend=self._backend,
            palette=palette,
        )
        # Now backend axes/viewbox are created, we can install mouse events
        # canvas._install_mouse_events()

        # link axes if needed
        if self._x_linked:
            self._x_linker_ref.link(canvas.x)
        if self._y_linked:
            self._y_linker_ref.link(canvas.y)
        canvas.events.drawn.connect(self.events.drawn.emit, unique=True, max_args=None)
        return canvas

    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for row in range(len(self._heights)):
            for col in range(len(self._widths)):
                yield self.add_canvas(row, col, **kwargs)

    def _iter_canvas(self) -> Iterator[tuple[tuple[int, int], Canvas]]:
        yielded: set[int] = set()
        for idx, canvas in np.ndenumerate(self._canvas_array):
            _id = id(canvas)
            if canvas is None or _id in yielded:
                continue
            yield idx, canvas
            yielded.add(_id)

    def show(self, block=False) -> None:
        """Show the grid."""
        from whitecanvas.backend._app import get_app

        # TODO: implement other event loops
        app = get_app(self._backend._app)
        _backend_app = app.get_app()
        out = self._backend_object._plt_show()

        if out is NotImplemented:
            from whitecanvas.backend._window import view

            view(self, self._backend.app)

        if block:
            # TODO: automatically block the event loop or enable ipython
            # GUI mode if needed.
            app.run_app()

    @property
    def background_color(self) -> NDArray[np.floating]:
        """Background color of the canvas."""
        return arr_color(self._backend_object._plt_get_background_color())

    @background_color.setter
    def background_color(self, color):
        self._backend_object._plt_set_background_color(arr_color(color))

    def screenshot(self) -> NDArray[np.uint8]:
        """Return a screenshot of the grid."""
        return self._backend_object._plt_screenshot()

    @property
    def size(self) -> tuple[int, int]:
        """Size in width x height."""
        return self._size

    @size.setter
    def size(self, size: tuple[int, int]):
        w, h = size
        if w <= 0 or h <= 0:
            raise ValueError("Size must be positive")
        self._size = (int(w), int(h))
        self._backend_object._plt_set_figsize(*self._size)

    @classmethod
    def from_dict(cls, d: dict[str, Any], backend: Backend | str | None = None) -> Self:
        _type_expected = f"{cls.__module__}.{cls.__name__}"
        if (_type := d.get("type")) and _type != _type_expected:
            raise ValueError(f"Expected type {_type_expected!r}, got {_type!r}")
        self = cls(d["heights"], d["widths"], backend=backend)
        if "size" in d:
            self.size = d["size"]
        if "background_color" in d:
            self.background_color = d["background_color"]
        for canvas_dict in d["canvas_array"]:
            if canvas_dict["canvas"]["type"].endswith("3D"):
                canvas = self.add_canvas_3d(canvas_dict["row"], canvas_dict["col"])
            else:
                canvas = self.add_canvas(canvas_dict["row"], canvas_dict["col"])
            canvas._update_from_dict(canvas_dict["canvas"])
        return self

    def to_dict(self) -> dict[str, Any]:
        return {
            "type": f"{type(self).__module__}.{type(self).__name__}",
            "heights": self._heights,
            "widths": self._widths,
            "size": self.size,
            "background_color": Color(self.background_color).hex,
            "canvas_array": [
                {"row": r, "col": c, "canvas": canvas.to_dict()}
                for (r, c), canvas in self._iter_canvas()
            ],
        }

    def copy(self, backend: Backend | str | None = None) -> Self:
        """Make a copy of the canvas."""
        if backend is None:
            backend = self._backend
        return self.from_dict(self.to_dict(), backend=backend)

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        from io import BytesIO

        try:
            from imageio import imwrite
        except ImportError:
            return None

        rendered = self.screenshot()
        if rendered is not None:
            with BytesIO() as file_obj:
                imwrite(file_obj, rendered, format="png")
                file_obj.seek(0)
                return file_obj.read()
        return None

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        if hasattr(self._backend_object, "_ipython_display_"):
            return self._backend_object._ipython_display_(*args, **kwargs)
        raise NotImplementedError()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        if hasattr(self._backend_object, "_repr_mimebundle_"):
            return self._backend_object._repr_mimebundle_(*args, **kwargs)
        raise NotImplementedError()

    def _repr_html_(self, *args: Any, **kwargs: Any) -> str:
        if hasattr(self._backend_object, "_repr_html_"):
            return self._backend_object._repr_html_(*args, **kwargs)
        raise NotImplementedError()

    def to_html(self, file: str | None = None) -> str:
        """Return HTML representation of the grid."""
        html = self._backend.get("to_html")(self._backend_object)
        if file is not None:
            Path(file).write_text(html, encoding="utf-8")
        return html


class CanvasVGrid(CanvasGrid):
    @classmethod
    def _from_heights(
        cls,
        heights: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> CanvasVGrid:
        return CanvasVGrid(heights, [1], backend=backend)

    @override
    def __getitem__(self, key: int) -> Canvas:
        canvas = self._canvas_array[key, 0]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        return canvas

    @override
    def add_canvas(self, row: int, **kwargs) -> Canvas:
        return super().add_canvas(row, 0, **kwargs)

    @override
    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for row in range(len(self._heights)):
            yield self.add_canvas(row, **kwargs)


class CanvasHGrid(CanvasGrid):
    @classmethod
    def _from_widths(
        cls,
        widths: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> CanvasHGrid:
        return CanvasHGrid([1], widths, backend=backend)

    @override
    def __getitem__(self, key: int) -> Canvas:
        canvas = self._canvas_array[0, key]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        return canvas

    @override
    def add_canvas(self, col: int, **kwargs) -> Canvas:
        return super().add_canvas(0, col, **kwargs)

    @override
    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for col in range(len(self._widths)):
            yield self.add_canvas(col, **kwargs)


class _CanvasWithGrid(CanvasBase):
    def __init__(self, canvas: Canvas, grid: CanvasGrid):
        self._main_canvas = canvas
        self._grid = grid
        super().__init__(palette=canvas._color_palette)

    @property
    def native(self) -> Any:
        """The native backend object."""
        return self._main_canvas.native

    def show(self, block: bool = False) -> None:
        """Show the canvas using the method defined in the backend."""
        self._grid.show(block=block)

    @property
    def background_color(self) -> NDArray[np.floating]:
        """Background color of the canvas."""
        return self._grid.background_color

    @background_color.setter
    def background_color(self, color):
        self._grid.background_color = color

    @property
    def size(self) -> tuple[float, float]:
        """Size of the canvas"""
        return self._grid.size

    @size.setter
    def size(self, size: tuple[float, float]):
        self._grid.size = size

    def screenshot(self) -> NDArray[np.uint8]:
        """Return a screenshot of the grid."""
        return self._grid.screenshot()

    def _get_backend(self) -> Backend:
        """Return the backend."""
        return self._main_canvas._backend

    def _canvas(self):
        return self._main_canvas._backend_object

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        return self._grid._repr_png_()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        return self._grid._repr_mimebundle_(*args, **kwargs)

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        return self._grid._ipython_display_(*args, **kwargs)

    def _repr_html_(self, *args: Any, **kwargs: Any) -> str:
        return self._grid._repr_html_(*args, **kwargs)

    def to_html(self, file: str | None = None) -> str:
        """Return HTML representation of the canvas."""
        return self._grid.to_html(file=file)


class SingleCanvas(_CanvasWithGrid, _Serializable):
    """
    A canvas without other subplots.

    This class is the simplest form of canvas. In `matplotlib` terms, it is a figure
    with a single axes.
    """

    def __init__(self, canvas: Canvas, grid: CanvasGrid):
        super().__init__(canvas, grid)

        # NOTE: events, dims etc are not shared between the main canvas and the
        # SingleCanvas instance. To avoid confusion, the first and the only canvas
        # should be replaces with the SingleCanvas instance.
        self.mouse = grid[0, 0].mouse
        grid._canvas_array[0, 0] = self
        self.events.drawn.connect(
            self._main_canvas.events.drawn.emit, unique=True, max_args=None
        )

    @classmethod
    def _new(cls, palette=None, backend=None) -> SingleCanvas:
        _grid = CanvasGrid([1], [1], backend=backend)
        _grid.add_canvas(0, 0, palette=palette)
        return SingleCanvas._from_grid(_grid)

    @classmethod
    def _from_grid(cls, grid: CanvasGrid) -> SingleCanvas:
        if grid.shape != (1, 1):
            raise ValueError(f"Grid shape must be (1, 1), got {grid.shape}")
        _it = grid._iter_canvas()
        _, canvas = next(_it)
        if next(_it, None) is not None:
            raise ValueError("Grid must have only one canvas")
        return cls(canvas, grid)

    @classmethod
    def from_dict(cls, d: dict[str, Any], backend: Backend | str | None = None) -> Self:
        """Create a SingleCanvas instance from a dictionary."""
        _type_expected = f"{cls.__module__}.{cls.__name__}"
        if (_type := d.get("type")) and _type != _type_expected:
            raise ValueError(f"Expected type {_type_expected!r}, got {_type!r}")
        self = cls._new(backend=backend, palette=d.get("palette"))
        self.layers.clear()
        self.layers.extend(construct_layers(d["layers"], backend=backend))
        self.x.update(d.get("x", {}))
        self.y.update(d.get("y", {}))
        self.title.update(d.get("title", {}))
        return self

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the canvas."""
        return {
            "type": f"{self.__module__}.{self.__class__.__name__}",
            "palette": self._color_palette,
            "layers": [layer.to_dict() for layer in self.layers],
            "title": self.title.to_dict(),
            "x": self.x.to_dict(),
            "y": self.y.to_dict(),
        }

    def copy(self, backend: Backend | str | None = None) -> Self:
        """Make a copy of the canvas."""
        if backend is None:
            backend = self._main_canvas._backend
        return self.from_dict(self.to_dict(), backend=backend)
